/*
 * milo/identify.c -- Identify the system we're running on
 *
 * Copyright (C) 1994, 1995 by Waldorf Electronics,
 * written by Andreas Busse and Ralf Baechle
 * Copyright (C) 1996 by Stoned Elipot
 * Modified by Stoned Elipot to support the tags
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License. See the file COPYING in the main directory of this archive for more
 * details.
 */
#include <linux/string.h>

#include <asm/bootinfo.h>
#include <asm/mipsregs.h>
#include <asm/segment.h>

#include <arc/arc.h>
#include <stand/stdio.h>
#include <stand/identify.h>
#include <stand/debug.h>
#include <stand/machspec.h>	/* machines specific routines */

extern machine_funcp mach_init;	/* from milo.c */

/*
 * Prototypes
 */
unsigned long decodecputype(int id);
char *cpuname(int type);

unsigned long val_icache_size     = 0; /* FIXME: cachectl.S needs these */
unsigned long val_icache_linesize = 0;
unsigned long val_dcache_size     = 0;
unsigned long val_dcache_linesize = 0;
unsigned long val_scache_size     = 0;
unsigned long val_scache_linesize = 0;

extern unsigned long mach_mem_upper; /* from milo.c */

/* values for tags of Mips Magnum 4000 and Olivetti M700-10 */
unsigned long machtype_MIPS_MAGNUM_4000 = MACH_MIPS_MAGNUM_4000;
unsigned long vram_base_MIPS_MAGNUM_4000 = 0xe0800000; 

/* values for tags of Deskstation rPC44 and Tyne */
unsigned long machtype_DESKSTATION_RPC44 = MACH_DESKSTATION_RPC44;
unsigned long machtype_DESKSTATION_TYNE = MACH_DESKSTATION_TYNE;
unsigned long vram_base_DESKSTATION_TYNE = 0xe10a0000; 

/* values for tags of Acer Pica 61 */
unsigned long machtype_ACER_PICA_61 = MACH_ACER_PICA_61;
unsigned long vram_base_ACER_PICA_61 = 0xe08b8000; 	/* for 80x25, 640x480 */

/* values for tags of unknown machine */
unsigned long machtype_UNKNOWN = MACH_UNKNOWN;
unsigned long machgroup_UNKNOWN = MACH_GROUP_UNKNOWN;

/* values for tags of ARC JAZZ machines */
unsigned long machgroup_JAZZ = MACH_GROUP_JAZZ;

/* misc values for tags */
unsigned long tlb_entries_48 = 48;
struct drive_info_struct drive_info_DUMMY = {{0,}}; 
unsigned long unsigned_long_KSEG0 = KSEG0;
unsigned long unsigned_long_KSEG0_8Mo = KSEG0ADDR(0x800000);
unsigned long unsigned_long_512x1024 = 512*1024;

/* zero values of various size for tags */
unsigned char unsigned_char_ZERO = 0;
unsigned int unsigned_int_ZERO = 0;
unsigned long unsigned_long_ZERO = 0;

/* these add clearity to the arrays below */
/* Hum,... or not ?? ;) -Stoned. */
#undef TAG
#undef TAGVAL
#define TAG(t,l)  {tag_##t,(l)}
#define TAGVAL(v) (void*)&(v)

/* tag list for Mips Magnum 4000 and Olivetti M700-10 */
tag_def taglist_mips_magnum_4000[] = {
  {TAG(machtype, ULONGSIZE), TAGVAL(machtype_MIPS_MAGNUM_4000)},
  {TAG(memlower, ULONGSIZE), TAGVAL(unsigned_long_KSEG0)},
  {TAG(machgroup, ULONGSIZE), TAGVAL(machgroup_JAZZ)},
  /*  {TAG(memupper, ULONGSIZE), TAGVAL(unsigned_long_KSEG0)},*/
  {TAG(scache_size, ULONGSIZE), TAGVAL(unsigned_long_ZERO)},
  {TAG(tlb_entries, ULONGSIZE), TAGVAL(tlb_entries_48)},
  /*  {TAG(dma_cache_size, ULONGSIZE), TAGVAL(unsigned_long_ZERO)},*/
  /*  {TAG(dma_cache_base, ULONGSIZE), TAGVAL(unsigned_long_ZERO)},*/
  {TAG(drive_info, DRVINFOSIZE), TAGVAL(drive_info_DUMMY)},
  {TAG(vram_base, ULONGSIZE), TAGVAL(vram_base_MIPS_MAGNUM_4000)},
  {TAG(dummy, 0), NULL}  /* must be the last one */
};

/* tag list for Deskstation rPC44 */
  /*
   * RB: This entry is absolutly untested - I don't have
   * even a hint about how the L2 cache works, so I make
   * the guess that it works just like the Tyne board.
   */
tag_def taglist_deskstation_rpc44[] = {
  /* WARNING ! memupper tag as to be the first in this list 
   * as long as we don't find a way to discover ourselves 
   * how much RAM there is in such a box 
   */
  {TAG(memupper, ULONGSIZE), TAGVAL(unsigned_long_KSEG0_8Mo)},
  {TAG(machtype, ULONGSIZE), TAGVAL(machtype_DESKSTATION_RPC44)},
  {TAG(machgroup, ULONGSIZE), TAGVAL(machgroup_UNKNOWN)},
  {TAG(memlower, ULONGSIZE), TAGVAL(unsigned_long_KSEG0)},
  {TAG(scache_size, ULONGSIZE), TAGVAL(unsigned_long_512x1024)},
  {TAG(tlb_entries, ULONGSIZE), TAGVAL(tlb_entries_48)},
  {TAG(dma_cache_size, ULONGSIZE), TAGVAL(unsigned_long_512x1024)},
  {TAG(dma_cache_base, ULONGSIZE), TAGVAL(unsigned_long_ZERO)},
  {TAG(drive_info, DRVINFOSIZE), TAGVAL(drive_info_DUMMY)},
  {TAG(vram_base, ULONGSIZE), TAGVAL(vram_base_DESKSTATION_TYNE)},
  {TAG(dummy, 0), NULL}  /* must be the last one */
};

/* tag list for Deskstation Tyne */
tag_def taglist_deskstation_tyne[] = {
  /* WARNING ! see the comment in the definition of 
   * taglist_deskstation_rpc44[] above about memupper tag
   */
  {TAG(memupper, ULONGSIZE), TAGVAL(unsigned_long_KSEG0_8Mo)},
  {TAG(machtype, ULONGSIZE), TAGVAL(machtype_DESKSTATION_TYNE)},
  {TAG(machgroup, ULONGSIZE), TAGVAL(machgroup_UNKNOWN)},
  {TAG(memlower, ULONGSIZE), TAGVAL(unsigned_long_KSEG0)},
  {TAG(scache_size, ULONGSIZE), TAGVAL(unsigned_long_512x1024)},
  {TAG(tlb_entries, ULONGSIZE), TAGVAL(tlb_entries_48)},
  {TAG(dma_cache_size, ULONGSIZE), TAGVAL(unsigned_long_512x1024)},
  {TAG(dma_cache_base, ULONGSIZE), TAGVAL(unsigned_long_ZERO)},
  {TAG(drive_info, DRVINFOSIZE), TAGVAL(drive_info_DUMMY)},
  {TAG(vram_base, ULONGSIZE), TAGVAL(vram_base_DESKSTATION_TYNE)},
  {TAG(dummy, 0), NULL}  /* must be the last one */
};

/* tag list fir Acer Pica 61 */
tag_def taglist_acer_pica_61[] = {
  {TAG(machtype, ULONGSIZE), TAGVAL(machtype_ACER_PICA_61)},
  {TAG(machgroup, ULONGSIZE), TAGVAL(machgroup_JAZZ)},
  {TAG(memlower, ULONGSIZE), TAGVAL(unsigned_long_KSEG0)},
  /*  {TAG(memupper, ULONGSIZE), TAGVAL(unsigned_long_KSEG0)},*/
  {TAG(scache_size, ULONGSIZE), TAGVAL(unsigned_long_512x1024)},
  {TAG(tlb_entries, ULONGSIZE), TAGVAL(tlb_entries_48)},
  /*  {TAG(dma_cache_size, ULONGSIZE), TAGVAL(unsigned_long_ZERO)},*/
  /*  {TAG(dma_cache_base, ULONGSIZE), TAGVAL(unsigned_long_ZERO)},*/
  {TAG(drive_info, DRVINFOSIZE), TAGVAL(drive_info_DUMMY)},
  {TAG(vram_base, ULONGSIZE), TAGVAL(vram_base_ACER_PICA_61)},
  {TAG(dummy, 0), NULL}  /* must be the last one */
};

/* tag list for unknown machine 
 * You don't really hope that Linux boot on such a system, don't you?
 *
 * Nb: commented tags are tags for which value is found out
 *     by identify() or set at runtime
 */
tag_def taglist_unknown[] = {
  /* WARNING ! memupper tag as to be the first tag here */
  {TAG(memupper, ULONGSIZE), TAGVAL(unsigned_long_KSEG0_8Mo)},
  {TAG(machtype, ULONGSIZE), TAGVAL(machtype_UNKNOWN)},
  {TAG(machgroup, ULONGSIZE), TAGVAL(machgroup_UNKNOWN)},
  /*  {TAG(cputype, ULONGSIZE), TAGVAL()},*/ /* detected */
  {TAG(memlower, ULONGSIZE), TAGVAL(unsigned_long_KSEG0)},
  /*  {TAG(icache_size, ULONGZISE), TAGVAL()},*/ /* detected */
  /*  {TAG(icache_linesize, ULONGZISE), TAGVAL()},*/ /* detected */
  /*  {TAG(dcache_size, ULONGZISE), TAGVAL()},*/ /* detected */
  /*  {TAG(dcache_linesize, ULONGZISE), TAGVAL()},*/ /* detected */
  {TAG(scache_size, ULONGSIZE), TAGVAL(unsigned_long_ZERO)},/* should be detected */
  /*  {TAG(scache_linesize, ULONGSIZE), TAGVAL()},*/ /* how this one is set/detected ?? */
  {TAG(tlb_entries, ULONGSIZE), TAGVAL(tlb_entries_48)}, /* true for all R4x00 except R4300i (32) */
  {TAG(dma_cache_size, ULONGSIZE), TAGVAL(unsigned_long_ZERO)},
  {TAG(dma_cache_base, ULONGSIZE), TAGVAL(unsigned_long_ZERO)},
  /* {TAG(ramdisk_size, ULONGSIZE), TAGVAL()},*/ /* runtime set */
  /* {TAG(ramdisk_base, ULONGSIZE), TAGVAL()},*/ /* runtime set */
  /*  {TAG(mount_root_rdonly, ULONGSIZE), TAGVAL()},*/ /* runtime set */
  {TAG(drive_info, DRVINFOSIZE), TAGVAL(drive_info_DUMMY)},
  {TAG(vram_base, ULONGSIZE), TAGVAL(unsigned_long_ZERO)},
  /*  {TAG(command_line, CMDLINESIZE), TAGVAL()},*/ /* runtime set - no support yet */
  {TAG(dummy, 0), NULL}  /* must be the last one */
};


/*
 *  Structure containing infos to get Milo identify
 *  a machine.
 */
struct machine_struct {
  char* SystemId;		/* same as value returned by getSystemId()  */
  
  char* VendorId;		/* VendorId is the same as the value        */
				/* returned by getVendorId() or (char*)NULL */
  
  char* identify_string;	/* string printed by identify() about the machine */
  
  machine_func* machine_ident;	/* specific system identification routine */
				/* (see for Acer Pica 61) */
  
  machine_func* machine_init;	/* specific system initialization before kernel */
				/* launch (see for DeskStation Tyne) */
  tag_def* taglist;		/* array of default tag to create */
};


/*
 * Array containing all the machine_struct corresponding to
 * known machines
 *
 * BEWARE!!! Order is revelant. See below in identify() on 
 *           how to add entries 
 */
struct machine_struct machinetbl[] = {
  { "DESKTECH-ARCStation I", (char*)NULL, "DeskStation rPC44", 
    (machine_func*)NULL, (machine_func*)NULL,
    taglist_deskstation_rpc44
  },

  { "DESKTECH-TYNE", (char*)NULL, "DeskStation Tyne",
    (machine_func*)NULL, deskstation_tyne_init,
    taglist_deskstation_tyne
  },

  { "PICA-61", (char*)NULL, "ACER PICA-61", 
    acer_pica_61_bi, (machine_func*)NULL, 
    taglist_acer_pica_61
  }, 

  /* This one is a Magnum: it's spelled O L I V E T T I */
  /* but pronounced Magnum :) */
  { "Microsoft-Jazz", "OLI00000", "Olivetti M700-10 (Jazz)",
    magnum_type_bi, magnum_type_init,
    taglist_mips_magnum_4000    
  },

  /* Someone please give me the vendor id string for MIPS */
  /* Thanks -stoned */
  { "Microsoft-Jazz", (char*)NULL, "Mips Magnum 4000 (Jazz)",
    magnum_type_bi, magnum_type_init,
    taglist_mips_magnum_4000    
  },

  /*
   * must be the last element.
   */
  { (char*)NULL, (char*)NULL, " Not identified -- using default parameters", 
    (machine_func*)NULL, (machine_func*)NULL, 
    taglist_unknown
  }
};

/*
 * identify()
 */
void identify(void)
{
  char *VendorId;		/* ARC vendor ID */
  char *ProductId;		/* ARC product ID */
  char *SystemId;		/* ARC system ID */
  int  prid;
  unsigned long type;
  int  config;
  int i;
  int machine_idx = 0;		/* index in machinetbl[] corresponding */
				/* to the one identified */
  
  /*
   * Get system information from ARC BIOS 
   */
  VendorId = getVendorId();
  ProductId = getProductId();
  printf("ARC Vendor  ID          : [");
  for(i=0; i<8; i++)
    printf("%c",(unsigned char)VendorId[i]);
  printf("]\r\n");
  printf("ARC Product ID          : [");
  for(i=0; i<8; i++)
    printf("%02x",(unsigned char)ProductId[i]);
  printf("]\r\n");
  SystemId = getSystemId();
  printf("ARC System  ID          : [%s]\r\n",SystemId);

  /* How machinetbl[] is being scanned? We search for a matching SystemId
   * and VendorId fields or for a matching SystemId and a null VendorId
   * So in machinetbl[], elements like SystemId = "XXX" VendorId="YYY"
   * *should* come before elements like SystemId = "XXX" VendorId = NULL
   * So specific distinctions are possible and a generic entry for each
   * SystemId can be used. See for Olivetti M700 and Magnum 4000.  No match
   * at all permits use of the default entry.
   */
  for(i=0;i<(sizeof(machinetbl)/sizeof(machinetbl[0]));i++)
  {
      /* BE CAREFULL: the order of this tests is important!
       * A NULL pointer shouldn't be passed to strncmp() or be prepared for
       * TLBE. So tests on NULL pointers are done before calls to strncmp().
       */
      if ((machinetbl[i].SystemId == (char*)NULL)
	  || ((!strncmp(machinetbl[i].SystemId, SystemId, 8))
	      && ((machinetbl[i].VendorId == (char*)NULL)
		  || (!strncmp(machinetbl[i].VendorId, VendorId, 8))))) 
      {
	  machine_idx = i;
	  break;	      
      }
  }

  /* special machine identification stuff ? 
   * if any, call it early, we may need it for finding out the 
   * memory upper limit 
   */
  if (machinetbl[machine_idx].machine_ident != (machine_func*)NULL)
    machinetbl[machine_idx].machine_ident(); 

  bi_TagAddList(machinetbl[machine_idx].taglist);

  /*
   * Get CPU Info
   */
  printf("ARC CPU ID              : [%s]\r\n",getCpuType());
  prid = read_32bit_cp0_register(CP0_PRID);
  type = decodecputype(prid);
  printf("Processor type          : [%s I%02x V%d.%d]\r\n",
	 cpuname(type),
	 prid >> 8, (prid & 0xf0) >> 4,prid & 0x0f);

  bi_TagAdd(tag_cputype, ULONGSIZE, (void*)&type);

  /*
   * Get cache and cache line sizes
   */
  config = read_32bit_cp0_register(CP0_CONFIG);
  val_icache_size = 1 << (12 + ((config & 0xe00) >> 9));
  val_dcache_size = 1 << (12 + ((config & 0x1c0) >> 6));
  val_icache_linesize = 16 << ((config & 0x20) >> 5);
  val_dcache_linesize = 16 << ((config & 0x10) >> 4);

  bi_TagAdd(tag_icache_size, ULONGSIZE, (void*)&val_icache_size);
  bi_TagAdd(tag_icache_linesize, ULONGSIZE, (void*)&val_icache_linesize);
  bi_TagAdd(tag_dcache_size, ULONGSIZE, (void*)&val_dcache_size);
  bi_TagAdd(tag_dcache_linesize, ULONGSIZE, (void*)&val_dcache_linesize);
  
  printf("Primary ICache size     : [0x%04x bytes]\r\n",val_icache_size);
  printf("Primary ICache line size: [0x%04x bytes]\r\n",val_icache_linesize);
  printf("Primary DCache size     : [0x%04x bytes]\r\n",val_dcache_size);
  printf("Primary DCache line size: [0x%04x bytes]\r\n",val_dcache_linesize);

  val_scache_size =  getSecondaryCacheSize();
  printf("Secondary cache size    : ");
  if (val_scache_size != SIZE_UNKNOWN)
      printf("[0x%4lx]\r\n", val_scache_size);
  else
      printf("[unknown]\r\n");

  /* For now the secondary cache size is set from the sample */
  /* taglists */
  /*  bi_TagAdd(tag_scache_size, ULONGSIZE, (void*)&val_scache_size);*/

  /* set the function pointer for specific machine initialisation */
  mach_init = machinetbl[machine_idx].machine_init;

  printf("ARC system identified as: %s\r\n\r\n", machinetbl[machine_idx].identify_string);
}

/*
 * cputype()
 * Extract cpu type from Processor Id
 * NOTE: This is not complete. Fill in here anything you know about.
 *
 * ID is organized as follows:
 *   Bits 31:16   always 0
 *   Bits 15:08   CPU Implementation
 *   Bits 07:04   CPU Version
 *   Bits 03:00   CPU Revision
 *
 * FIXME: CPU dependent stuff should go into a global array.
 */
unsigned long decodecputype(int id)
{
  switch (id >> 8)
    {
    case 0x01:
      return CPU_R2000;

    case 0x02:
      return CPU_R3000;

    case 0x03:
      return CPU_R6000;
      
    case 0x04:			/* any of R4000xx and R4400xx */
      switch (id & 0xf0)
	{
	case 0x00:		/* R4000 */
	case 0x10:
	case 0x20:
	case 0x30:
	  return CPU_R4000PC;
	  
	case 0x40:		/* R4400 V4.x with daddiu bug */
	case 0x50:		/* R4400 V5.x */
	  return CPU_R4400PC;
	  
	default:
	  return CPU_UNKNOWN;
	}
      break;

    case 0x05:			/* LSI Logic R3000 derivative */
      return CPU_R3000;

    case 0x06:
      return CPU_R6000A;

    case 0x07:
      return CPU_R3051;		/* IDT 3051,3052 etc. */

    case 0x09:
      return CPU_R10000;

    case 0x0a:
      return CPU_R4200;

    case 0x10:
      return CPU_R8000;

    case 0x20:
      return CPU_R4600;		/* R4600 V1.x has an icache bug! */

    case 0x21:
      return CPU_R3000;		/* Sony R3000 derivative */

    case 0x22:
      return CPU_R3000;		/* Toshiba R3000 derivate */

    case 0x23:
      return CPU_R3000;		/* NKK R3000 derivative */
      
    default:
      return CPU_UNKNOWN;
      
    }
}

static char *cpu_name_strings[] = CPU_NAMES;
/*
 * cpuname()
 * Translate cputype into a processor name
 * FIXME: This should also go into a global array of cpu informations.
 *        Some of these CPUs are too stupid (no MMU) to run Linux.
 */
char *cpuname(int type)
{
    if(type > CPU_LAST)
        type = CPU_UNKNOWN;

    return cpu_name_strings[type];
}
