#ifndef __BOOT_H
#define __BOOT_H

#include <arc/types.h>

/*
 * Boot status
 */
typedef struct _BOOT_STATUS {
    ULONG BootStarted : 1;
    ULONG BootFinished : 1;
    ULONG RestartStarted : 1;
    ULONG RestartFinished : 1;
    ULONG PowerFailStarted : 1;
    ULONG PowerFailFinished : 1;
    ULONG ProcessorReady : 1;
    ULONG ProcessorRunning : 1;
    ULONG ProcessorStart : 1;
} BOOT_STATUS, *PBOOT_STATUS;

/*
 * Restart Block
 */
#define RESTART_BLOCK_MAGIC 0x42545352

typedef struct _RESTART_BLOCK {
    ULONG RSTBSignature;
    ULONG RSTBLength;
    USHORT Version;
    USHORT Revision;
    struct _RESTART_BLOCK *NextRestartBlock;
    PVOID RestartAddress;
    ULONG BootMasterId;
    ULONG ProcessorId;
    volatile BOOT_STATUS BootStatus;
    ULONG CheckSum;
    ULONG SaveAreaLength;
    ULONG SaveArea[1];
} RESTART_BLOCK, *PRESTART_BLOCK;
    
/*
 * Debug Block
 * (Currently neither defined nor used for Windows NT)
 */
typedef struct _DEBUG_BLOCK {
} DEBUG_BLOCK, *PDEBUG_BLOCK;

/*
 * System parameter block, located at 0x1000
 *
 * Note: The Microsoft "NT Portable Boot Loader Specification"
 *       contains the SYSTEM_PARAMETER_BLOCK_MAGIC erroneously
 *       specified as 0x53435241.
 */
#define SYSTEM_PARAMETER_BLOCK_MAGIC 0x41524353

typedef struct _SYSTEM_PARAMETER_BLOCK {
    ULONG Signature;
    ULONG Length;
    USHORT Version;
    USHORT Revision;
    PRESTART_BLOCK RestartBlock;
    PDEBUG_BLOCK DebugBlock;
    PVOID GeneralExceptionVector;
    PVOID TlbMissExceptionVector;
    ULONG FirmwareVectorLength;
    PVOID *FirmwareVector;
    ULONG VendorVectorLength;
    PVOID *VendorVector;
    ULONG AdapterCount;
    ULONG Adapter0Type;
    ULONG Adapter0Length;
    PVOID *Adapter0Vector;
} SYSTEM_PARAMETER_BLOCK, *PSYSTEM_PARAMETER_BLOCK;

typedef struct _SYSTEM_ID {
    CHAR VendorId[8];
    CHAR ProductId[8];
} SYSTEM_ID, *PSYSTEM_ID;

#endif /* __BOOT_H */
