#ifndef __CONFIGURATION_H
#define __CONFIGURATION_H

#include <arc/types.h>

/*
 * System information supplied by the ARCS BIOS.
 */
typedef struct {
  CHAR VendorId[8];
  CHAR ProductId[8];
} SYSTEMID;

typedef enum _CONFIGURATION_CLASS {
    SystemClass,
    ProcessorClass,
    CacheClass,
    AdapterClass,
    ControllerClass,
    PeripheralClass,
    MemoryClass
} CONFIGURATION_CLASS, *PCONFIGURATION_CLASS;

typedef enum _CONFIGURATION_TYPE {
    System,

    CentralProcessor,
    FloatingPointProcessor,

    PrimaryIcache,
    PrimaryDcache,
    SecondaryIcache,
    SecondaryDcache,
    SecondaryCache,

    EisaAdapter,                 /* Eisa adapter         */
    TcAdapter,                   /* Turbochannel adapter */
    ScsiAdapter,                 /* SCSI adapter         */
    DtiAdapter,                  /* AccessBus adapter    */
    MultiFunctionAdapter,

    DiskController,
    TapeController,
    CdromController,
    WormController,
    SerialController,
    NetworkController,
    DisplayController,
    ParallelController,
    PointerController,
    KeyboardController,
    AudioController,
    OtherController,             /* denotes a controller not otherwise defined */

    DiskPeripheral,
    FloppyDiskPeripheral,
    TapePeripheral,
    ModemPeripheral,
    MonitorPeripheral,
    PrinterPeripheral,
    PointerPeripheral,
    KeyboardPeripheral,
    TerminalPeripheral,
    OtherPeripheral,           /* denotes a peripheral not otherwise defined   */
    LinePeripheral,
    NetworkPeripheral,

    SystemMemory
} CONFIGURATION_TYPE, *PCONFIGURATION_TYPE;

typedef struct _DEVICE_FLAGS {
    ULONG Failed : 1;
    ULONG ReadOnly : 1;
    ULONG Removeable : 1;
    ULONG ConsoleIn : 1;
    ULONG ConsoleOut : 1;
    ULONG Input : 1;
    ULONG Output : 1;
} DEVICE_FLAGS, *PDEVICE_FLAGS;

typedef struct _CONFIGURATION_COMPONENT {
    CONFIGURATION_CLASS Class;
    CONFIGURATION_TYPE Type;
    DEVICE_FLAGS Flags;
    USHORT Version;
    USHORT Revision;
    ULONG Key;
    ULONG AffinityMask;
    ULONG ConfigurationDataLength;
    ULONG IdentifierLength;
    PCHAR Identifier;
} CONFIGURATION_COMPONENT, * PCONFIGURATION_COMPONENT;

typedef struct _CM_PARTIAL_RESOURCE_DESCRIPTOR {
    UCHAR Type;
    UCHAR ShareDisposition;
    USHORT Flags;
    union {
        struct {
            PHYSICAL_ADDRESS Start;
            ULONG Length;
            } Port;

        struct {
            ULONG Level;
            ULONG Vector;
            ULONG Affinity;
            } Interrupt;

        struct {
            PHYSICAL_ADDRESS Start;
            ULONG Length;
            } Memory;

        struct {
            ULONG Channel;
            ULONG Port;
            ULONG Reserved1;
            } Dma;

        struct {
            ULONG DataSize;
            ULONG Reserved1;
            ULONG RESERVED2;
            } DeviceSpecificData;
        } u;
} CM_PARTIAL_RESOURCE_DESCRIPTOR, *PCM_PARTIAL_RESOURCE_DESCRIPTOR;

typedef struct _CM_PARTIAL_RESOURCE_LIST {
    USHORT Version;
    USHORT Revision;
    ULONG Count ;
    CM_PARTIAL_RESOURCE_DESCRIPTOR Partial_Descriptors[1];
} CM_PARTIAL_RESOURCE_LIST, *PCM_PARTIAL_RESOURCE_LIST;

#endif /* __CONFIGURATION_H */
