/*
 * Copyright (C) 1995 TERMA Elektronik A/S, Lystrup, Denmark
 * Written by Kai Harrekilde-Petersen <kah@terma.dk>
 *
 * Version: 0.32
 * Date: 7. July 1995
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <unistd.h>
#if defined(DISASM_STDALONE)
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#define VERSION "0.32"
#else
#include <stand/behave.h>
#include <stand/stdio.h>
#define DISASM_VERSION "0.32 (NAME_OF_MILO/NAME_OF_PANDORA version)"
#endif /* DISASM_STDALONE */

/* instruction classes */
#define TYPE_NONE -1 /* opcode is invalid */
#define TYPE_NOP 0
#define TYPE_IMMU 1
#define TYPE_REG 2
#define TYPE_JUMP 3
#define TYPE_IMM 4
#define TYPE_BRA 5
#define TYPE_LDST 6
#define TYPE_SPECIAL 7
#define TYPE_REGIMM 8
#define TYPE_SHIFT 9
#define TYPE_JR 10
#define TYPE_SYSCALL 11
#define TYPE_HILO 12
#define TYPE_MOVE 13
#define TYPE_CP 14

/*
 * Instruction types:
 * NOP:		nop
 * IMMU:	andi	r1,r2,unsigned16
 * REG:		or	r1,r2,r3
 * JUMP:	j	imm28
 * IMM:		addi	r1,r2,signed16
 * BRA:		beq	r1,r2,`signed16 << 2'
 * LDST:	lw	r1,signed16(r2)
 * REGIMM:	ori	r1,r2,signed16 
 * SHIFT:	sll	r1,r2,unsigned16
 * JR:		jr	r1
 * SYSCALL:	syscall	unsigned20
 * HILO:	mtlo	r1
 * MOVE:	move	r1,r2
 * SPECIAL:	any of: SHIFT, REG, JR, SYSCALL, HILO, and MOVE
 * CP:		any of: NOP, REG
 */

typedef struct {
  char *name;
  int type;
} insn;

#define FIELD(insn,shift,width) ((insn >> shift) & (~((~0) << width)))

void
disassemble(char *, unsigned, unsigned);


insn opcode_list[64]= {
  {"", TYPE_SPECIAL}, {"", TYPE_REGIMM},
  {"j", TYPE_JUMP}, {"jal", TYPE_JUMP},
  {"beq", TYPE_BRA}, {"bne", TYPE_BRA},
  {"blez", TYPE_BRA}, {"bgtz", TYPE_BRA},
  {"addi", TYPE_IMM}, {"addiu", TYPE_IMM},
  {"slti", TYPE_IMMU}, {"sltiu", TYPE_IMMU},
  {"andi", TYPE_IMMU}, {"ori", TYPE_IMMU},
  {"xori", TYPE_IMMU}, {"lui", TYPE_IMMU},
  {"cop0", TYPE_CP}, {"cop1", TYPE_CP},
  {"cop2", TYPE_CP}, {"cop3", TYPE_CP},
  {"beql", TYPE_BRA}, {"bnel", TYPE_BRA},
  {"blezl", TYPE_BRA}, {"bgtzl", TYPE_BRA},
  {"daddi", TYPE_IMM}, {"daddiu", TYPE_IMM},
  {"ldl", TYPE_LDST}, {"ldr", TYPE_LDST},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"lb", TYPE_LDST}, {"lh", TYPE_LDST},
  {"lwl", TYPE_LDST}, {"lw", TYPE_LDST},
  {"lbu", TYPE_LDST}, {"lhu", TYPE_LDST},
  {"lwr", TYPE_LDST}, {"lwu", TYPE_LDST},
  {"sb", TYPE_LDST}, {"sh", TYPE_LDST},
  {"swl", TYPE_LDST}, {"sw", TYPE_LDST},
  {"sdl", TYPE_LDST}, {"sdr", TYPE_LDST},
  {"swr", TYPE_LDST}, {"cache", TYPE_NONE},
  {"ll", TYPE_LDST}, {"lwc1", TYPE_LDST},
  {"lwc2", TYPE_LDST}, {"lwc3", TYPE_LDST},
  {"lld", TYPE_LDST}, {"ldc1", TYPE_LDST},
  {"ldc2", TYPE_LDST}, {"ld", TYPE_LDST},
  {"sc", TYPE_LDST}, {"swc1", TYPE_LDST},
  {"swc2", TYPE_LDST}, {"swc3", TYPE_LDST},
  {"scd", TYPE_LDST}, {"sdc1", TYPE_LDST},
  {"sdc2", TYPE_LDST}, {"sd", TYPE_LDST}
};

insn special_list[64]= {
  {"sll",TYPE_SHIFT}, {"",TYPE_NONE},
  {"srl",TYPE_SHIFT}, {"sra",TYPE_SHIFT},
  {"sllv",TYPE_REG}, {"",TYPE_NONE},
  {"srlv",TYPE_REG}, {"srav",TYPE_REG},
  {"jr",TYPE_JR}, {"jalr",TYPE_JR},
  {"",TYPE_NONE}, {"",TYPE_NONE},
  {"syscall",TYPE_SYSCALL}, {"break",TYPE_SYSCALL},
  {"",TYPE_NONE}, {"sync",TYPE_NOP},
  {"mfhi",TYPE_HILO}, {"mthi",TYPE_HILO},
  {"mflo",TYPE_HILO}, {"mtlo",TYPE_HILO},
  {"dsslv",TYPE_REG}, {"",TYPE_NONE},
  {"dsrlv",TYPE_REG}, {"dsrav",TYPE_REG},
  {"mult",TYPE_MOVE}, {"multu",TYPE_MOVE},
  {"div",TYPE_MOVE}, {"divu",TYPE_MOVE},
  {"dmult",TYPE_MOVE}, {"dmultu",TYPE_MOVE},
  {"ddiv",TYPE_MOVE}, {"ddivu",TYPE_MOVE},
  {"add",TYPE_REG}, {"addu",TYPE_REG},
  {"sub",TYPE_REG}, {"sub",TYPE_REG},
  {"and",TYPE_REG}, {"or",TYPE_REG},
  {"xor",TYPE_REG}, {"nor",TYPE_REG},
  {"",TYPE_NONE}, {"",TYPE_NONE},
  {"slt",TYPE_REG}, {"sltu",TYPE_REG},
  {"dadd",TYPE_REG}, {"daddu",TYPE_REG},
  {"dsub",TYPE_REG}, {"dsubu",TYPE_REG},
  {"tge",TYPE_MOVE}, {"tgeu",TYPE_MOVE},
  {"tlt",TYPE_MOVE}, {"tltu",TYPE_MOVE},
  {"teq",TYPE_MOVE}, {"",TYPE_NONE},
  {"tne",TYPE_MOVE}, {"",TYPE_NONE},
  {"dsll",TYPE_SHIFT}, {"",TYPE_NONE},
  {"dsrl",TYPE_SHIFT}, {"dsra",TYPE_SHIFT},
  {"dsll32",TYPE_SHIFT}, {"",TYPE_NONE},
  {"dsrl32",TYPE_SHIFT}, {"dsra32",TYPE_SHIFT}
};

insn regimm_list[32]= {
  {"bltz", TYPE_BRA}, {"bgez", TYPE_BRA},
  {"bltzl", TYPE_BRA}, {"bgezl", TYPE_BRA},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"tgei", TYPE_MOVE}, {"tgeiu", TYPE_MOVE},
  {"tlti", TYPE_MOVE}, {"tltiu", TYPE_MOVE},
  {"teqi", TYPE_MOVE}, {"", TYPE_NONE},
  {"tnei", TYPE_MOVE}, {"", TYPE_NONE},
  {"bltzal", TYPE_BRA}, {"bgezal", TYPE_BRA},
  {"bltzall", TYPE_BRA}, {"bgezall", TYPE_BRA},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE}
};

insn special_regimm_list[32]= {
  {"bltz", TYPE_BRA}, {"b", TYPE_BRA},
  {"bltzl", TYPE_BRA}, {"bl", TYPE_BRA},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"bltzal", TYPE_BRA}, {"bal", TYPE_BRA},
  {"bltzall", TYPE_BRA}, {"ball", TYPE_BRA},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE}
};

insn COPzrs_list[32] = {
  {"mf", TYPE_CP}, {"dmf", TYPE_CP},
  {"cf", TYPE_CP}, {"", TYPE_NONE},
  {"mt", TYPE_CP}, {"dmt", TYPE_CP},
  {"ct", TYPE_CP}, {"", TYPE_NONE},
  {"bc", TYPE_CP}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"co", TYPE_NOP}, {"co", TYPE_NOP},
  {"co", TYPE_NOP}, {"co", TYPE_NOP},
  {"co", TYPE_NOP}, {"co", TYPE_NOP},
  {"co", TYPE_NOP}, {"co", TYPE_NOP},
  {"co", TYPE_NOP}, {"co", TYPE_NOP},
  {"co", TYPE_NOP}, {"co", TYPE_NOP},
  {"co", TYPE_NOP}, {"co", TYPE_NOP},
  {"co", TYPE_NOP}, {"co", TYPE_NOP}
};
  
insn COPzrt_list[32] = {
  {"bcf", TYPE_CP}, {"bctf", TYPE_CP},
  {"bcfl", TYPE_CP}, {"bctl", TYPE_CP},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE}
};
  
insn CP0_list[64] = {
  {"", TYPE_NONE}, {"tlbr", TYPE_NOP}, {"tlbwi", TYPE_NOP}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE}, {"tlbwr", TYPE_NOP}, {"", TYPE_NONE},
  {"tlbp", TYPE_NOP}, {"", TYPE_NONE},  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"eret", TYPE_NOP}, {"", TYPE_NONE},  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"wait", TYPE_NOP}, {"", TYPE_NONE},  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},  {"", TYPE_NONE}, {"", TYPE_NONE}
};

insn CP1_func[64] = {
  {"add", TYPE_REG}, {"sub", TYPE_REG},
  {"mul", TYPE_REG}, {"div", TYPE_REG},
  {"sqrt", TYPE_MOVE}, {"abs", TYPE_MOVE},
  {"mov", TYPE_MOVE}, {"neg", TYPE_MOVE},
  {"round.l", TYPE_MOVE}, {"trunc.l", TYPE_MOVE},
  {"ceil.l", TYPE_MOVE}, {"floor.l", TYPE_MOVE},
  {"round.w", TYPE_MOVE}, {"trunc.w", TYPE_MOVE},
  {"ceil.w", TYPE_MOVE}, {"floor.w", TYPE_MOVE},
  /* 0x10 .. 0x1f is reserved */
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"cvt.s", TYPE_MOVE}, {"cvt.d", TYPE_MOVE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"cvt.w", TYPE_MOVE}, {"cvt.l", TYPE_MOVE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"c.f", TYPE_BRA}, {"c.un", TYPE_BRA},
  {"c.eq", TYPE_BRA}, {"c.ueq", TYPE_BRA},
  {"c.olt", TYPE_BRA}, {"c.ult", TYPE_BRA},
  {"c.ole", TYPE_BRA}, {"c.ule", TYPE_BRA},
  {"c.sf", TYPE_BRA}, {"c.ngle", TYPE_BRA},
  {"c.seq", TYPE_BRA}, {"c.ngl", TYPE_BRA},
  {"c.lt", TYPE_BRA}, {"c.nge", TYPE_BRA},
  {"c.le", TYPE_BRA}, {"c.ngt", TYPE_BRA}
};

insn CP1_fmt[16] = {
  {"s", TYPE_NOP}, {"d", TYPE_NOP},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"w", TYPE_NOP}, {"l", TYPE_NOP},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE},
  {"", TYPE_NONE}, {"", TYPE_NONE}
};

char pfx[2];

unsigned char *regis[32] = {
  "zero", "at", "v0", "v1",
  "a0",   "a1", "a2", "a3",
  "t0",   "t1", "t2", "t3",
  "t4",   "t5", "t6", "t7",
  "s0",   "s1", "s2", "s3",
  "s4",   "s5", "s6", "s7",
  "t8",   "t9", "k1", "k2",
  "gp",   "sp", "s8", "ra"
};

unsigned char *CP0regs[32] = {
  "index",	"random",	"entrylo",	"",
  "context",	"pagemask",	"wired",	"error",
  "bvaddr",	"count",	"entryhi",	"compare",
  "sr",		"cause",	"epc",		"prid",
  "config",	"lladdr",	"watchlo",	"watchhi",
  "",		"",		"",		"",
  "",		"",		"ecc",		"cacheerr",
  "taglo",	"taghi",	"errepc",	""
};


void
print_special(char *string, unsigned insn, unsigned addr)
{
  unsigned opcode;
  unsigned r1, r2, r3;
  unsigned shift;
  unsigned imm16;
  unsigned imm26;

  signed int imm;

  /*
   * Well, the *real* opcode field lies at [31..26], but we look at
   * the sub-opcode here (main opcode is "special").
   */
  opcode = FIELD(insn,0,6);
  r1 = FIELD(insn,21,5);
  r2 = FIELD(insn,16,5);
  r3 = FIELD(insn,11,5);
  imm16 = FIELD(insn,0,16);
  imm26 = FIELD(insn,0,26);
  shift = FIELD(insn,6,5);

  switch(special_list[opcode].type) {
  case TYPE_REG:
    sprintf(string,"%s\t%s%s,%s%s,%s%s", special_list[opcode].name,pfx,
	    regis[r3],pfx,regis[r1],pfx,regis[r2]);
    break;
  case TYPE_IMM:
    /* convert imm16 into a "true" signed int */
    imm = (imm16 & (1 << 15)) ? -(1 << 16) + imm16 : imm16;
    if(imm >= 0 ) {
      sprintf(string,"%s\t%s%s,%s%s,0x%04x",special_list[opcode].name,pfx,
	      regis[r2],pfx,regis[r1],imm);
    } else {
      sprintf(string,"%s\t%s%s,%s%s,%i",special_list[opcode].name,pfx,
	      regis[r2],pfx,regis[r1],imm);
    }
    break;
  case TYPE_SHIFT:
    sprintf(string,"%s\t%s%s,%s%s,0x%02x",special_list[opcode].name,pfx,
	    regis[r3],pfx,regis[r2],shift);
    break;
  case TYPE_JR:
    sprintf(string,"%s\t%s%s",special_list[opcode].name,pfx,regis[r1]);
    break;
  case TYPE_SYSCALL:
    /* extract "code" field */
    imm = FIELD(insn,6,20);
    sprintf(string,"%s\t0x%05x",special_list[opcode].name,imm);
    break;
  case TYPE_NOP:
    sprintf(string,"%s",special_list[opcode].name);
    break;
  case TYPE_HILO:
    sprintf(string,"%s\t%s%s",special_list[opcode].name,pfx,regis[r2]);
    break;
  case TYPE_MOVE:
    sprintf(string,"%s\t%s%s,%s%s",opcode_list[opcode].name,pfx,regis[r1],
	    pfx,regis[r2]);
    break;
  case TYPE_NONE:
    /* illegal insn */
    sprintf(string,"illegal\t# print_special():TYPE_NONE");
    break;
  default:
    /* Internal error -- cannot happen! */
    sprintf(string,"Oops!\tinternal error in print_special().");
    break;
  }
}

/*
 * parse a COPz insn
 */
void
print_cp(char *string, unsigned insn, unsigned addr)
{
  unsigned opcode;
  unsigned cp;
  unsigned r1, r2, r3, r4;

  /*
   * Well, the *real* opcode field lies at [31..26], but we look at
   * the sub-opcode here (main opcode is "COPz").
   */
  opcode = FIELD(insn,0,6);
  cp = FIELD(insn,26,2);
  r1 = FIELD(insn,21,5);
  r2 = FIELD(insn,16,5);
  r3 = FIELD(insn,11,5);
  r4 = FIELD(insn,6,5);

  switch(cp) {
  case 0: /* System co-proc */
    switch(COPzrs_list[r1].type) {
    case TYPE_CP:
      sprintf(string,"%sc0\t%s%s,%s%s",COPzrs_list[r1].name,pfx,regis[r2],
	      pfx,CP0regs[r3]);
      break;
    case TYPE_NOP:
      /* TYPE_NOP: check sub-opcode for tlbxx, eret, wait insns */
      switch(CP0_list[opcode].type) {
      case TYPE_NOP:
	sprintf(string,"%s",CP0_list[opcode].name);
	break;
      case TYPE_NONE:
	sprintf(string,"cp0\t0x%02x",opcode);
	break;
      default: /* This cannot happen */
	sprintf(string,"print_cp(): This cannot happen! "
		"opcode = 0x%02x",opcode);
	break;
      } /* switch(CP0_list[opcode].type) */
    default:
      sprintf(string,"illegal\t# print_cp():CP0_list:default");
      break;
    } /* switch(COPzrs_list) */
    break;
    /* END OF CP0 PROCESSING */
  case 1: /* Floating Point co-proc */
    if(opcode == 0x00) {
      /* "standard" move from/to CP1 insns */
      switch(COPzrs_list[r1].type) {
      case TYPE_CP:
	sprintf(string,"%sc1\t%s%s,%sf%i",COPzrs_list[r1].name,pfx,
		regis[r2],pfx,r3);
	break;
      case TYPE_NOP:
	/* Naughty, naughty:  --FIXME!!!! */ 
	if(r1 != 16 && r1 != 17) {
	  /* illegal format specifier */
	  sprintf(string,"%s.ill\t%sf%i,%sf%i",CP1_func[opcode].name,pfx,
		  r3,pfx,r2);
	} else {
	  sprintf(string,"%s.%s\t%sf%i,%sf%i",CP1_func[opcode].name,
		  CP1_fmt[r1].name,pfx,r3,pfx,r2);
	}
	break;
      case TYPE_NONE:
	sprintf(string,".word\t0x%08x",insn);
	break;
      default: /* This cannot happen */
	sprintf(string,"print_cp(): This cannot happen! "
		"opcode = 0x%02x",opcode);
	break;
      } /* switch(CP0_list[opcode].type) */
    } else {
      /* "real" FPU insns */
      switch(CP1_func[opcode].type) {
      case TYPE_REG:
	if(r1<16 || CP1_fmt[r1-16].type == TYPE_NONE) {
	  /* illegal format specifier */
	  sprintf(string,"%s.ill\t%sf%i,%sf%i,%sf%i",
		  CP1_func[opcode].name,pfx,r4,pfx,r3,pfx,r2);
	} else {
	  sprintf(string,"%s.%s\t%sf%i,%sf%i,%sf%i",CP1_func[opcode].name,
		  CP1_fmt[r1].name,pfx,r4,pfx,r3,pfx,r2);
	}
	break;
      case TYPE_MOVE:
	if(r1<16 || CP1_fmt[r1-16].type == TYPE_NONE) {
	  /* illegal format specifier */
	  sprintf(string,"%s.ill\t%sf%i,%sf%i",CP1_func[opcode].name,pfx,
		  r4,pfx,r3);
	} else {
	  sprintf(string,"%s.%s\t%sf%i,%sf%i,%sf%i",CP1_func[opcode].name,
		  CP1_fmt[r1].name,pfx,r4,pfx,r3,pfx,r2);
	}
	break;
      case TYPE_BRA:
	/* This is *compares* not branches(!) */
	/* C.cond.fmt $fA,$fB */
	if(r1<16 || CP1_fmt[r1-16].type == TYPE_NONE) {
	  /* illegal format specifier */
	  sprintf(string,"%s.ill\t%sf%i,%sf%i",CP1_func[opcode].name,pfx,
		  r3,pfx,r2);
	} else {
	  sprintf(string,"%s.%s\t%sf%i,%sf%i",CP1_func[opcode].name,
		  CP1_fmt[r1].name,pfx,r3,pfx,r2);
	}
	break;
      case TYPE_NONE:
	sprintf(string,"illegal\t# print_cp():CP1_func:TYPE_NONE");
	break;
      default:
	sprintf(string,"\t# Oops! internal error in print_cp():CP1_func)");
	break;
      } /* switch(CP1_func[opcode].type) */
    } /* !if(opcode == 0x00) */
  case 2: /* ??? co-proc */
    sprintf(string,"cp%i\t# Not implemented yet!",cp);
    break;
  case 3: /* invalid?? co-proc */
    sprintf(string,"cp3: illegal coproc");
    break;
  default: /* cannot happen */
    sprintf(string,"print_cp(): This cannot happen. cp = %i",cp);
    break;
  }
  return;
}

/*
 * examine the machine instruction `insn' and format it into `string',
 * under the assumption that it resides at `addr'.
 */
void
disassemble(char *string, unsigned insn, unsigned addr)
{
  unsigned opcode;
  unsigned r1, r2, r3;
  unsigned shift;
  unsigned imm16;
  unsigned imm26;
  unsigned offset;
  signed int imm;
  unsigned long target;

  opcode = FIELD(insn,26,6);
  r1 = FIELD(insn,21,5);
  r2 = FIELD(insn,16,5);
  r3 = FIELD(insn,11,5);
  imm16 = FIELD(insn,0,16);
  imm26 = FIELD(insn,0,26);
  offset = FIELD(insn,0,6);
  shift = FIELD(insn,6,6);

  /*
   * Handle all "special cases" first
   */
  switch(opcode_list[opcode].type) {
  case TYPE_SPECIAL:
    /* "sll $0,$0,0" => "nop" */
    if(insn == 0x00) {
      sprintf(string,"nop");
      return;
    }
    /* map "or $rA,$zero,$rB" => "move $rA,$rB" */
    if(offset == 0x25 && r2 == 0x00) {
      sprintf(string,"move\t%s%s,%s%s",pfx,regis[r3],pfx,regis[r1]);
      return;
    }
    break;
  case TYPE_REGIMM:
    if(regimm_list[r2].type == TYPE_BRA && r1 == 0) {
      imm = (imm16 & (1 << 15)) ? -(1<<18) + (imm16 << 2) : imm16 << 2;
      target = (unsigned long)(addr + 4 + imm);
      if(special_regimm_list[r2].type == TYPE_NONE) {
	sprintf(string,"\t# disassemble():SpecialCases(TYPE_REGIMM):"
		" This cannot happen");
      } else {
	sprintf(string,"%s\t%08lx",special_regimm_list[r2].name,target);
      }
      return;
    } /* if(regimm_list[r2].type == TYPE_BRA && r1 == 0) */
    break;
  case TYPE_IMMU:
    /* "lui $rA,$zero,imm" => "lui $rA,imm" */
    if(opcode == 0x0f && r1 == 0x00) {
      sprintf(string,"lui\t%s%s,0x%04x",pfx,regis[r2],imm16);
      return;
    }
    break;
  case TYPE_IMM:
    /* "addiu $rA,$zero,imm" => "li $rA,imm" */
    if(opcode == 0x09 && r1 == 0x00) {
      imm = (imm16 & (1 << 15)) ? -(1 << 16) + imm16 : imm16;
      if(imm >= 0) {
	sprintf(string,"li\t%s%s,0x%04x",pfx,regis[r2],imm);
      } else {
	sprintf(string,"li\t%s%s,%i",pfx,regis[r2],imm);
      }
      return;
    }
    break;
  case TYPE_BRA:
    /* "beq $0,$0,target" => "b target" */
    if(r1 == 0 && r2 == 0) {
      /* shift imm16 two bits left and signed extend it */
      imm = (imm16 & (1 << 15)) ? -(1<<18) + (imm16 << 2) : imm16 << 2;
      /* compute target address */
      target = (unsigned long)(addr + 4 + imm);
      sprintf(string,"b\t%08lx",target);
      return;
    }
    /* "bCC $r1,$0,target" => "bCCz $r1,target" */
    if(r1 == 0 || r2 == 0) {
      /* shift imm16 two bits left and signed extend it */
      imm = (imm16 & (1 << 15)) ? -(1<<18) + (imm16 << 2) : imm16 << 2;
      /* compute target address */
      target = (unsigned long)(addr + 4 + imm);
      if(r1 == 0) r1 = r2; /* Cover both cases graciously */
      sprintf(string,"%sz\t%s%s,%08lx",opcode_list[opcode].name,pfx,
	      regis[r1],target);
      return;
    }
  case TYPE_CP:
    if(COPzrs_list[r1].type == TYPE_CP && FIELD(insn,26,2) == 0
       && (r1 == 0x02 || r1 == 0x06)) {
      /* CFCz and CTCz are illegal for COP0 */
      sprintf(string,"illegal\t# disassemble():Special cases:TYPE_CP");
      return;
    }
    break;
  } /* Handle "special cases" */

  /*
   * Now handle the normal cases
   */
  switch(opcode_list[opcode].type) {
  case TYPE_SPECIAL:
    print_special(string,insn,addr);
    break;
  case TYPE_REGIMM:
    imm = (imm16 & (1<<15)) ? -(1<<16) + imm16 : imm16;
    switch(regimm_list[r2].type) {
    case TYPE_BRA:
      target = addr + 4 + (imm << 2);
      sprintf(string,"%s\t%s%s,%08lx",regimm_list[r2].name,pfx,
	      regis[r1],target);
      break;
    case TYPE_MOVE:
      if(imm < 0) {
	sprintf(string,"%s\t%s%s,%i",regimm_list[r2].name,pfx,
		regis[r1],imm);
      } else {
	sprintf(string,"%s\t%s%s,0x%04x",regimm_list[r2].name,pfx,
		regis[r1],imm);
      }
      break;
    case TYPE_NONE:
      sprintf(string,".word\t0x%08x\t# Illegal instruction",insn);
      break;
    default:
      sprintf(string,"disassemble():switch(regimm_list): "
	      "This cannot happen");
      break;
    } /* switch(regimm_list[r2].type) */
    break;
  case TYPE_IMMU:
    sprintf(string,"%s\t%s%s,%s%s,0x%04x",opcode_list[opcode].name,pfx,
	    regis[r2],pfx,regis[r1],imm16);
    break;
  case TYPE_JUMP:
    target = (addr & (~0 << 28)) | (imm26 << 2);
    sprintf(string,"%s\t%08lx",opcode_list[opcode].name,target);
    break;
  case TYPE_IMM:
    imm = (imm16 & (1 << 15)) ? -(1 << 16) + imm16 : imm16;
    if(imm >= 0) {
      sprintf(string,"%s\t%s%s,%s%s,0x%04x",opcode_list[opcode].name,pfx, 
	      regis[r2],pfx,regis[r1],imm);
    } else {
      sprintf(string,"%s\t%s%s,%s%s,%i",opcode_list[opcode].name,pfx,
	      regis[r2],pfx,regis[r1],imm);
    }
    break;
  case TYPE_BRA:
    /* shift imm16 two bits left and signed extend it */
    imm = (imm16 & (1 << 15)) ? -(1<<18) + (imm16 << 2) : imm16 << 2;
    /* compute target address */
    target = (unsigned long)(addr + 4 + imm);
    sprintf(string,"%s\t%s%s,%s%s,%08lx",opcode_list[opcode].name,pfx,
	    regis[r1],pfx,regis[r2],target);
    break;
  case TYPE_LDST:
    imm = (imm16 & (1 << 15)) ? -(1 << 16) + imm16 : imm16;
    sprintf(string,"%s\t%s%s,0x%x(%s%s)",opcode_list[opcode].name,pfx,
	    regis[r2],imm,pfx,regis[r1]);
    break;
  case TYPE_JR:
    sprintf(string,"%s\t%s%s",opcode_list[opcode].name,pfx,regis[r1]);
    break;
  case TYPE_CP:
    print_cp(string,insn,addr);
    break;
  case TYPE_NONE:
    sprintf(string,"illegal\t# disassemble():Normal cases:TYPE_NONE");
    break;
  case TYPE_NOP:
  case TYPE_REG:
  case TYPE_SHIFT:
  case TYPE_SYSCALL:
  case TYPE_HILO:
  case TYPE_MOVE:
    /* These can only happen as a sub-opcode under SPECIAL */
  default:
    /* illegal insn or something wrong */
    sprintf(string,"Oops!\tinternal error in disassemble(): "
	    "opcode type = %03x",opcode_list[opcode].type );
    break;
  }
}

/*
 * convert a big/little endian word into a little/big endian word
 */
inline unsigned long
nuxi (const unsigned long value)
{
  return ((value >> 24) & 0x000000ffUL)
       | ((value >>  8) & 0x0000ff00UL)
       | ((value <<  8) & 0x00ff0000UL)
       | ((value << 24) & 0xff000000UL);
}

#if defined(DISASM_STDALONE)
void
main(int argc, char **argv)
{
  u_long dummy_l;
  char stri[1025];
  char *binname;
  int flag;
  int reverse;
  unsigned addr;
  int counter;
  FILE *stream1;
  struct stat file_s;

  /*
   * DEFAULTS:
   * No little/big endian conversion
   * Start address = 0xbfc0000
   * Emit Algorithmics style reg names ($rx, as opposed to mips style: rx)
   */
  reverse = 0;
  addr = 0xbfc00000;
  pfx[0] = '$';
  pfx[1] = '\0';

  /* find exactly the "basename" of argv[0] */
  binname = strrchr(argv[0],'/');
  if(binname == NULL) {
    binname = argv[0];
  } else {
    binname++;
  }

  if(argc == 1) {  
    printf("%s : [-hprvV] [-a <start address>] <filename>\n",binname);
    exit(0);
  }

  /* initialize optind for getopt() */
  optind = 1;
  do {
    flag = getopt(argc,argv,"a:hprvV");
    switch(flag) {
    case 'r': 
      reverse = 1;
      break;
    case 'a':
      addr = strtol(optarg,(char **)NULL,0);
      break;
    case 'h':
      printf("%s: [-hprvV] [-a <start address>] <filename>\n"
	     "-h:\tshow this help\n"
	     "-p:\temit Algorithmics style regs ('$rx' vs 'rx')\n"
	     "-r:\treverse the endianness\n"
	     "-v:\tshow version number\n"
	     "-V:\tshow copying information\n",binname);
      exit(0);
      break;
    case 'p':
      pfx[0] = '\0';
      break;
    case 'v':
      printf("%s version " VERSION "\n",binname);
      exit(0);
      break;
    case 'V':
      printf("%s: MIPS R4x00 disassembler, version " VERSION
	     " by Kai Harrekilde-Petersen.\n"
	     "Copyright (C) 1995 TERMA Elektronik A/S\n"
	     "This is free software; see the source for copying "
	     "conditions.\nThere is NO warranty; not even for"
	     "MERCHANTABILITY or FITNESS FOR A\nPARTICULAR PURPOSE.\n",
	     binname);
      exit(0);
      break;
    case ':':
    case '?':
      printf("I'm sorry. I don't understand `%s'\n",argv[optind]);
      exit(2);
    }
  } while(flag != -1);

  if(optind < argc) {
    flag=stat(argv[optind],&file_s);
    if (flag==-1) {
      printf("Couldn't stat file %s, aborting program !!\n",argv[optind]);
      exit(1);
    }
  } else {
    puts("Hmmm. I couldn't find a filename in there.");
    exit(0);
  }

/* Opening file */
  stream1=fopen(argv[optind],"r");
  strcpy(stri,argv[optind]);

  for(counter = 0; counter < file_s.st_size/4; counter++) {
    fread(&dummy_l,sizeof(unsigned long),1,stream1);
    if(reverse) dummy_l = nuxi(dummy_l);
    stri[0] = '\0'; /* clear the string buffer */
    disassemble(stri, dummy_l,addr);
    printf("    %08lx: %08lx\t",(unsigned long) addr,dummy_l);
    printf("%s\n",stri);
    addr += 4;
  }
  fclose(stream1);
  exit(0);
}
#endif /*  DISASM_STDALONE */
