/*
 * lib/dumpcp0r.c
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of this archive for
 * more details.
 *
 * Copyright (C) by Stoned Elipot <Stoned.Elipot@fnet.fr>
 */
#include <asm/bootinfo.h>
#include <asm/mipsregs.h>
#include <linux/string.h>

#include <stand/console.h>
#include <stand/debugger.h>
#include <stand/stdio.h>
#include <arc/arc.h>

static int position;

int decodecputype(int id);

static void
dump_32bit(char *format, unsigned long content)
{
    printf(format, content);
    if (position < 3)
    {
        position++;
        printf("\t");
    }
    else
    {
        position = 1;
        printf("\r\n");
    }
}

/*
 * Dump the CP0 registers
 * FIXME: This is based on the R4000 proc.
 *        Are there more or less registers on other Rxxxx ?
 */
int 
dump_cp0r(void)
{
    unsigned int type = decodecputype(read_32bit_cp0_register(CP0_PRID));
    position = 1;

    printf("CP0 registers\r\n");
    dump_32bit("$0  Index   : %08x", read_32bit_cp0_register(CP0_INDEX));
    /*
     * The random register is uninteresting
     *
    dump_32bit("$1  Random  : %08x", read_32bit_cp0_register(CP0_RANDOM));
     */
    dump_32bit("$2  EntryLo0: %08x", read_32bit_cp0_register(CP0_ENTRYLO0)); 
    dump_32bit("$3  EntryLo1: %08x", read_32bit_cp0_register(CP0_ENTRYLO1));
    dump_32bit("$4  Context : %08x", read_32bit_cp0_register(CP0_CONTEXT));
    dump_32bit("$5  PageMask: %08x", read_32bit_cp0_register(CP0_PAGEMASK)); 

    switch(type)
    {
    case CPU_R4000PC:
    case CPU_R4000SC:
    case CPU_R4000MC:
    case CPU_R4200:
    case CPU_R4400PC:
    case CPU_R4400SC:
    case CPU_R4400MC:
    case CPU_R4600:
    case CPU_R6000:
    case CPU_R6000A:
    case CPU_R8000:
    case CPU_R10000:
        dump_32bit("$6  Wired   : %08x", read_32bit_cp0_register(CP0_WIRED));
    }

    dump_32bit("$8  BadVAddr: %08x", read_32bit_cp0_register(CP0_BADVADDR));
    dump_32bit("$9  Count   : %08x", read_32bit_cp0_register(CP0_COUNT)); 
    dump_32bit("$10 EntryHi : %08x", read_32bit_cp0_register(CP0_ENTRYHI));
    dump_32bit("$11 Compare : %08x", read_32bit_cp0_register(CP0_COMPARE));
    dump_32bit("$12 Status  : %08x", read_32bit_cp0_register(CP0_STATUS));
    dump_32bit("$13 Cause   : %08x", read_32bit_cp0_register(CP0_CAUSE));
    dump_32bit("$14 EPC     : %08x", read_32bit_cp0_register(CP0_EPC));
    dump_32bit("$15 PRId    : %08x", read_32bit_cp0_register(CP0_PRID)); 

    switch(type)
    {
    case CPU_R4000PC:
    case CPU_R4000SC:
    case CPU_R4000MC:
    case CPU_R4200:
    case CPU_R4400PC:
    case CPU_R4400SC:
    case CPU_R4400MC:
    case CPU_R4600:
        dump_32bit("$16 Config  : %08x", read_32bit_cp0_register(CP0_CONFIG));
        break;
    }

    switch(type)
    {
    case CPU_R4000PC:
    case CPU_R4000SC:
    case CPU_R4000MC:
    case CPU_R4200:
    case CPU_R4400PC:
    case CPU_R4400SC:
    case CPU_R4400MC:
    case CPU_R4600:
    case CPU_R6000:
    case CPU_R6000A:
    case CPU_R8000:
    case CPU_R10000:
        dump_32bit("$17 LLAddr  : %08x", read_32bit_cp0_register(CP0_LLADDR));
        break;
    }

    switch(type)
    {
    case CPU_R4000PC:
    case CPU_R4000SC:
    case CPU_R4000MC:
    case CPU_R4400PC:
    case CPU_R4400SC:
    case CPU_R4400MC:
    /*
     * FIXME: Don't know about R4200, R6000, R6000A, R8000 and R10000
     */
        dump_32bit("$18 WatchLo : %08x", read_32bit_cp0_register(CP0_WATCHLO)); 
	dump_32bit("$19 WatchHi : %08x", read_32bit_cp0_register(CP0_WATCHHI));
        dump_32bit("$20 XContext: %08x", read_32bit_cp0_register(CP0_XCONTEXT));
        dump_32bit("$26 ECC     : %08x", read_32bit_cp0_register(CP0_ECC)); 
        dump_32bit("$27 CacheErr: %08x", read_32bit_cp0_register(CP0_CACHEERR));
        dump_32bit("$28 TagLo   : %08x", read_32bit_cp0_register(CP0_TAGLO));
        dump_32bit("$29 TagHi   : %08x", read_32bit_cp0_register(CP0_TAGHI)); 
        break;
    }

    dump_32bit("$30 ErrorEPC: %08x", read_32bit_cp0_register(CP0_ERROREPC));

    if (position > 1)
    {
        printf("\r\n");
    }

    return 0;
}
