/*
 * lib/dumptlb.c
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of this archive for
 * more details.
 *
 * Copyright (C) 1994, 1995 by Waldorf Electronics,
 * written by Ralf Baechle and Andreas Busse
 */
#include <asm/mipsregs.h>
#include <linux/string.h>
#include <sys/file.h>

#include <stand/console.h>
#include <stand/stdio.h>
#include <arc/arc.h>

static char temp[1024];

#define VALID 0x02ULL		/* valid bit */
#define TLBFILE "multi(0)disk(0)fdisk(0)\\TLBDUMP"

void
dump_tlb(int all, int con)
{
    int			i, wired, dummy;
    FILE		*fd;
    unsigned long long	entry_hi, entry_lo0, entry_lo1;

    if (!con)
    {
	fd = fopen(TLBFILE, "w");
	if (fd < 0)
	{
	    printf("dump_tlb(): could not open `%s'.\r\n",TLBFILE);
	    return;
	}
	printf("\r\n\r\ndump_tlb(): writing TLB dump to file TLBDUMP...");
	fflush(stdout);
    }
    else
	fd = stdout;
  
    fprintf(fd, "\r\n\r\nTLB Dump:\r\n\r\n");
    wired = read_32bit_cp0_register(CP0_WIRED);
    fprintf(fd, "Wired: %d\r\n\r\n", wired);
  
    for(i=0 ; i<48; i++)		/* print all TLB entries */
    {
	write_32bit_cp0_register(CP0_INDEX, i);
	__asm__ __volatile__(
		".set\tnoreorder\n\t"
		"nop;nop;nop;nop\n\t"
		"tlbr\n\t"
		"nop;nop;nop;nop\n\t"
		".set\treorder");
      
	entry_hi = read_64bit_cp0_register(CP0_ENTRYHI);
	entry_lo0 = read_64bit_cp0_register(CP0_ENTRYLO0);
	entry_lo1 = read_64bit_cp0_register(CP0_ENTRYLO1);
      
	if ((entry_lo0 && VALID) || (entry_lo1 && VALID) || all)
	{
	    fprintf(fd, "TLB #%d:\r\n",i);
	    fprintf(fd, "\tEntryHi : %08x:%08x\t",
		    (unsigned int)(entry_hi >> 32), (unsigned int) entry_hi);
	    fprintf(fd, "(R=%1x VADR=%02x:%08x ASID=%04x)\r\n",
		    entry_hi >> 62,
		    entry_hi & 0xff00000000ULL,
		    entry_hi & 0xffffe000ULL,
		    entry_hi & 0xffULL);
	    fprintf(fd, "\tEntryLo0: %08x:%08x\t",
		    (unsigned int)(entry_lo0 >> 32), (unsigned int) entry_lo0);
	    fprintf(fd, "(PADR=%02x:%08x C=%d D=%d V=%d G=%d)\r\n",
		    (entry_lo0 & 0x3c000000) >> 25,
		    (entry_lo0 & 0x03ffffc0) << 6,
		    (entry_lo0 & 0x00000038) >> 3,
		    (entry_lo0 & 0x00000004) >> 2,
		    (entry_lo0 & 0x00000002) >> 1,
		    (entry_lo0 & 0x00000001));
	    fprintf(fd, "\tEntryLo1: %08x:%08x\t",
		    (unsigned int)(entry_lo1 >> 32), (unsigned int) entry_lo1);
	    fprintf(fd, "(PADR=%02x:%08x C=%d D=%d V=%d G=%d)\r\n",
		    (entry_lo1 & 0x3c000000) >> 25,
		    (entry_lo1 & 0x03ffffc0) << 6,
		    (entry_lo1 & 0x00000038) >> 3,
		    (entry_lo1 & 0x00000004) >> 2,
		    (entry_lo1 & 0x00000002) >> 1,
		    (entry_lo1 & 0x00000001));
	    fprintf(fd, "\tPageMask: %08x\r\n\r\n",
		    read_32bit_cp0_register(CP0_PAGEMASK));
	    if (con)
	    {
		printf("Press a key for next\r\n");
		getchar();
	    }
	}
    }

    if (!con)
	close(fd);
  
    printf(" done.\r\n");
}
