/*
 * btn (better than nothing) malloc(3)
 *
 * FIXME:
 * - chained lists are slow
 * - calloc(3) and realloc(3) aren't supported
 */
#include <arc/configuration.h>
#include <arc/errno.h>
#include <arc/memory.h>
#include <arc/proto.h>
#include <sys/types.h>

#include <stand/stdio.h>
#include <stand/debug.h>
#include <stand/malloc.h>

static int malloc_is_initialised = 0;

void initmalloc(void);
void dumpmalloc(void);

static void remove_chunk_from_pool(struct mpool *p)
{
    struct mpool *pool, *previous;
    ULONG size;

    pool = _FirstChunk;
    previous = NULL;
    size = *(PULONG)p;
    while(pool) {
        if(pool == p) {
            if(previous)
                previous->NextChunk = p->NextChunk;
            else
                _FirstChunk = p->NextChunk;
            return;
            }
        previous = pool;
        pool = pool->NextChunk;
        }
}

static void add_chunk_to_pool(struct mpool *p, size_t size)
{
    struct mpool *pool, *previous;

    /*
     * try to find a chunk we can join the returned one with
     */
    pool = _FirstChunk;
    previous = NULL;
    while(pool) {
        if(((PVOID)pool) + pool->Size == p) {
            pool->Size += size;
            return;
            }
        else if (((PVOID)p) + size == pool) {
            if(previous)
                previous->NextChunk = p;
            else
                _FirstChunk = p;
            p->NextChunk = pool->NextChunk;
            p->Size = pool->Size + size;
            return;
            }
        previous = pool;
        pool = pool->NextChunk;
        }
    /*
     * Found no chunk to melt with, so add as a lone chunk
     */
    p->NextChunk = _FirstChunk;
    p->Size = size;
    _FirstChunk = p;
}

void *malloc(size_t size)
{
    struct mpool *p, *previous;

    if(!malloc_is_initialised)
    {
        initmalloc();
        malloc_is_initialised = 1;
    }
    if (debuglevel >= 2)
	dumpmalloc();

    size = (size + 1023) & ~1023;
    /*
     * reserve space for size word
     */
/*    size += sizeof(ULONG); */
    size += 0x10;

    /*
     * Search for suitable memory chunk
     */
    p = _FirstChunk;
    while(p) {
        if(size <= p->Size)
	{
            /*
             * Found a large enough piece of memory
             */
            remove_chunk_from_pool(p);
	    if (debuglevel >= 2)
		dumpmalloc();
            if(p->Size - size)
                add_chunk_to_pool(((PVOID)p) + size, p->Size - size);
            *(PULONG)p = size;
	    if (debuglevel >= 2)
		printf("MALLOC: %x, %d\r\n", ((PULONG)p) + 1, size);
/*            return ((PULONG)p) + 1; */
	    return ((PUCHAR)p)+0x10;
	}
        previous = p;
        p = p->NextChunk;
    }

    /*
     * Nothing found, return error
     */
    return NULL;
}

void free(void *p)
{
    /*
     * free(NULL) is legal
     */
    if(p == NULL)
	return;

    p -= sizeof(ULONG);
    add_chunk_to_pool(p, *(PULONG)p);
}

