/*
 * milo/mallocinit.c
 * btn (better than nothing) malloc(3)
 *
 * Copyright (C) 1994 by Waldorf Electronics,
 * written by Ralf Baechle and Andreas Busse
 *
 * FIXME:
 * - chained lists are slow
 * - calloc(3) and realloc(3) aren't supported
 */
#include <arc/configuration.h>
#include <arc/errno.h>
#include <arc/memory.h>
#include <arc/proto.h>
#include <sys/types.h>
#include <asm/segment.h>

#include <stand/malloc.h>

extern int _end;
extern int loadaddr;

struct mpool *_FirstChunk = NULL;
static inline int checkchunk(PMEMORY_DESCRIPTOR d)
{
  PVOID addr;

  addr = (PVOID)(d->BasePage << PAGE_SHIFT);

  /*
   * This function is a bugfix for the Deskstation Tyne ARC BIOS
   * which returns an memory descriptor for address
   * 0x84000000, size 0x400000. Maybe it's a BIOS bug in 1.3.017,
   * maybe it's just because I've got only 8mb installed which
   * conflicts with the specification.
   */

  if(addr == (PVOID) 0x04000000)
    return 0;
  if(d->MemoryType != MemoryFree)
    return 0;

  /*
   * Seems to be legal.
   */

  return 1;
}

void initmalloc(void)
{
  PMEMORY_DESCRIPTOR d = NULL;
  ULONG p, size, chunkend;

  while((d = GetMemoryDescriptor(d)))
    {
      /*
       * Add free RAM to list, do as if RAM is mapped to ksegx
       */

      if(checkchunk(d))
	{
	  p = (d->BasePage << PAGE_SHIFT) | KSEGX(loadaddr);
	  size = d->PageCount << PAGE_SHIFT;
	  if(p >= loadaddr && p < (ULONG)&_end)
	    {
	      chunkend = p + size;
	      p = ((ULONG)&_end) + (PAGESIZE-1);
	      p = p & ~(PAGESIZE-1);
	      size = chunkend - p;
	    }
	  ((struct mpool *)p)->NextChunk = _FirstChunk;
	  ((struct mpool *)p)->Size      = size;
	  _FirstChunk = (struct mpool *)p;
	}
    }
}

/* end of file mallocinit.c */
