#include <stand/stdio.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/file.h>

#include <arc/types.h>
#include <arc/errno.h>
#include <arc/pathname.h>
#include <stand/debug.h>
#include <arc/proto.h>

int open(const char *pathname, int flags, ...)
{
    OPEN_MODE arcmode;
    _STATUS st;
    ULONG fd;

    if (debuglevel >= 2)
    {
	printf("open(): entered with pathname=%s, flags=%d\r\n",
	       pathname, flags);
	fflush(stdout);
    }

    switch(flags & O_ACCMODE)
    {
    case O_RDONLY:
        arcmode = OpenReadOnly;
        break;
    case O_WRONLY:
        arcmode = (flags & O_CREAT) ? CreateWriteOnly : OpenWriteOnly;
        break;
    case O_RDWR:
        arcmode = (flags & O_CREAT) ? CreateReadWrite : OpenReadWrite;
        break;
    default:
        errno = EINVAL;
        return -1;
    }

    if (debuglevel >= 2)
    {
	printf("open(): Calling ARC Open(%s,%d,%p)\r\n",
	       pathname, arcmode, &fd);
	fflush(stdout);
    }
    st = Open(pathname, arcmode, &fd);
    if(st != ESUCCESS)
    {
        errno = st;
        return -1;
    }
    if (debuglevel >= 2)
    {
	printf("open(): ARC Open() returned fd=%d\r\n",fd);
	fflush(stdout);
    }
    return fd;
}
