#include <sys/types.h>
#include <unistd.h>

#include <asm/bootinfo.h>

#include <arc/arc.h>
#include <arc/types.h>
#include <arc/errno.h>
#include <arc/pathname.h>
#include <arc/proto.h>
#include <stand/debug.h>
#include <stand/stdio.h>

#define BLOCKSIZE  0x200	/* Mips Magnum 4000 BIOS crashes when */
				/* reading more than 512 bytes at once ! */

/*
 * Read from file in BLOCKSIZE bytes chunks. This workaround is way to slow
 * on other systems, so use it only for the broken Magnums.
 */
static __inline__ ssize_t splitted_read(int fd, __ptr_t buf, size_t count)
{
    ULONG Count;
    ULONG n;
    _STATUS st;
    ULONG c;

    Count = 0;
    while (count)
    {
	n = (count < BLOCKSIZE ? count : BLOCKSIZE);
	if (debuglevel >= 2)
	{
	    printf("splitted_read(): count=%d, n=%d, Count=%d\r\n",
		   count, n, Count);
	}
	st = Read(fd, (PVOID) buf, (ULONG) n, &c);
	if(st != ESUCCESS)
        {
	    errno = st;
	    return -1;
        }
	Count += c;
	buf += c;
	count -= c;
    }

    return (ssize_t) Count;
}

static __inline__ ssize_t default_read(int fd, __ptr_t buf, size_t count)
{
    ULONG Count;
    _STATUS st;

    if (debuglevel >= 2)
	printf("default_read(): count=%d\r\n", count);

    st = Read(fd, (PVOID) buf, (ULONG) count, &Count);
    if(st != ESUCCESS)
    {
        errno = st;
        return -1;
    }

    return (ssize_t) Count;
}

ssize_t read(int fd, __ptr_t buf, size_t count)
{
    ssize_t c;

    if(__splitted_rw)
	c = splitted_read(fd, buf, count);
    else
	c = default_read(fd, buf, count);

    return c;
}
