/*
 * lib/seek.c
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 1995 by Waldorf GMBH
 * written by Ralf Baechle
 */
#include <stand/stdio.h>
#include <sys/file.h>

#include <arc/types.h>
#include <arc/errno.h>
#include <arc/pathname.h>
#include <arc/proto.h>

off_t
lseek(int fd, off_t offset, int whence)
{
    LARGE_INTEGER off;
    _STATUS st;
    FILE_INFORMATION fi;

    off = offset;
    switch(whence)
    {
    case L_SET:
        st = Seek((ULONG) fd, &off, (SEEK_MODE) SeekAbsolute);
        if(st != ESUCCESS)
	{
            errno = st;
            return -1;
        }
        break;
    case L_INCR:
        st = Seek((ULONG) fd, &off, (SEEK_MODE) SeekRelative);
        if(st != ESUCCESS)
        {
            errno = st;
            return -1;
        }
        break;
    case L_XTND:
	st = GetFileInformation(fd, &fi);
	if(st != ESUCCESS)
	{
	    errno = st;
	    return -1;
	}
	return (off_t) fi.EndingAddress;
	break;
    default:
	errno = EINVAL;
	return -1;
    }
    st = GetFileInformation(fd, &fi);
    if(st != ESUCCESS)
    {
        errno = st;
        return -1;
    }

    return (off_t) fi.CurrentPosition;
}
