/*
 * milo/sleep.c
 *
 * Copyright (C) 1994, 1995 by Waldorf Electronics,
 * written by Ralf Baechle and Andreas Busse
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file README.legal in the main directory of this archive
 * for more details.
 */
#include <arc/types.h>
#include <arc/time.h>
#include <arc/proto.h>

/*
 * sleep()
 * Wait an arbitrary amount of time specified in seconds.
 * FIXME: This might fail shortly before midnight at the end of a month :-)
 */
unsigned int sleep(unsigned int seconds)
{
    TIMEINFO *current;
    TIMEINFO end;

    current = (TIMEINFO *)GetTime();		/* get current time */

    /* calculate time to reach */
  
    end.Hundredths = current->Hundredths;
    end.Seconds = current->Seconds + (seconds % 60);
    end.Minutes = current->Minutes + ((seconds / 60) % 60);
    end.Hour = current->Hour + ((seconds / 3600) % 24);
    end.Day = current->Day;
    end.Month = current->Month;
    end.Year = current->Year;
   
    /* check overflow */

    if (end.Seconds > 59)
    {
	end.Seconds %= 60;
	end.Minutes++;
    }
    if (end.Minutes > 59)
    {
	end.Minutes %= 60;
	end.Hour++;
    }
    if (end.Hour > 23)
    {
	end.Hour %= 24;
	end.Day++;
    }

    /* wait until time is reached */

    while(current->Day != end.Day ||
	  current->Hour != end.Hour ||
	  current->Minutes != end.Minutes ||
	  current->Seconds != end.Seconds)
	current = (TIMEINFO *)GetTime();

    return(0);
}

/*
 * usleep()
 * Sleep an arbitrary amount of time specified in microseconds.
 * NOTE: ARC GetTime() always returns Hundredths = 0,
 *       therefortime is quantized to one second.
 */
unsigned int usleep(unsigned int micros)
{
    return(sleep(micros/100000));
}
