/*
 * magnum.c
 * Magnum type specific routines:
 * used for MIPS Magnum 4000 and Olivetti M700-10
 *
 * Copyright (C) 1996 Stoned Elipot & Luc Beurton
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License. See the file COPYING in the main directory of this archive for more
 * details.
 */
#include <stand/stdio.h>
#include <linux/tty.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/bootinfo.h>
#include <stand/debug.h>
#include <asm/jazz.h>
#include <hdw/g364_def.h>

/* will be tags soon */
extern struct screen_info si;

/*
 * magnum_type_init()
 * Do video mode detection before launch the beast
 */
int magnum_type_init(void)
{
  unsigned int mx,my;
  unsigned long i;
  if (debuglevel >=2)
    printf("Magnum type pre-launch initialization\r\n");

  si.orig_x = 0;		/* start at col. 0, line 0 */
  si.orig_y = 0;
  G364_get_display(&mx,&my);
  si.orig_video_cols = mx / 8;
  si.orig_video_lines = my / 16;
  for(i=0;i<262143;i++)
	(unsigned long long )(*((unsigned long long*)G364_VRAM+i))=0x0404040404040404;

  return 0;
}

extern unsigned long mach_mem_upper; /* from identify.c */

#define MB (1024*1024)

/*
 * magnum_type_bi()
 * Detect the installed RAM size
 * See Olivetti's 'M700-10 RISC Personnal Computer - 
 * hardware-architecture and function' pp 3-4, 3-5
 */
int magnum_type_bi(void)
{
  unsigned int gcregv;
  unsigned char bankindice;
  unsigned int banksize=0;
  unsigned int totalram=0;
  int i;
  int bankenabled;
  int tweenbankenabled;


  gcregv = (unsigned int)(*((unsigned int *) JAZZ_R4030_CONFIG));
  for(i = 0; i<2; i++)
    {
      bankenabled =  ((int)((gcregv >> (4 + i)) & 0x1));
      if (bankenabled)
	{
	  tweenbankenabled = ((int)((gcregv >> (4 + i + 2)) & 0x1));
	  bankindice = ((unsigned char)((gcregv >> ((i % 2) ? 2 : 0)) & 0x3)); 
	  switch(bankindice)
	    {
	    case 0:
	      banksize = (tweenbankenabled ? 2 : 1);
	      break;
	    case 1:
	      banksize = (tweenbankenabled ? 8 : 4);
	      break;
	    case 2:
	      banksize = (tweenbankenabled ? 32 : 16);
	      break;
	    case 3:
	      banksize = 0;
	      break;
	    }  
	  totalram+= (banksize << 2) * MB;
	}
    }
  mach_mem_upper = KSEG0 + totalram;
  /* we don't need to create a memupper tag, but for consistency */
  bi_TagAdd(tag_memupper, ULONGSIZE, (void*)&mach_mem_upper);
  return 0;
}
